package ml2java.parser;

import java.io.Reader;
import java.io.FileReader;
import java.io.IOException;
import java.util.NoSuchElementException;
import java.util.Objects;

import ml2java.util.*;

import static ml2java.util.SimpleList.*;
import static ml2java.util.Pair.*;
import static ml2java.util.Option.*;
import static ml2java.util.Bind.*;


public abstract class Typ {
    
    public static Typ Int() {
        return new Int();
    }
    public static Typ Bool() {
        return new Bool();
    }
    
    public abstract <R> R match(Function0<R> f1, Function0<R> f2);
    
    public abstract <R> R accept(TypVisitor<R> v);
    
    public interface TypVisitor<R> {
    
        public R visitInt();
        public R visitBool();
    
    }
    public static abstract class TypDefaultVisitor<R> implements TypVisitor<R> {
    
        public R visitInt() {
            return visitDefault();
        }
        public R visitBool() {
            return visitDefault();
        }
        public abstract R visitDefault();
    
    }
}

class Int extends Typ { 
    public Int() {
    }
    @Override
    public int hashCode() {
        return Objects.hash();
    }
    @Override
    public boolean equals(Object other) {
        if (other == null) return false;
        return (this.getClass() == (other.getClass()));
    }
    public <R> R match(Function0<R> f1, Function0<R> f2) {
        return f1.apply();
    }
    public <R> R accept(TypVisitor<R> v) {
        return v.visitInt();
    }
}
class Bool extends Typ { 
    public Bool() {
    }
    @Override
    public int hashCode() {
        return Objects.hash();
    }
    @Override
    public boolean equals(Object other) {
        if (other == null) return false;
        return (this.getClass() == (other.getClass()));
    }
    public <R> R match(Function0<R> f1, Function0<R> f2) {
        return f2.apply();
    }
    public <R> R accept(TypVisitor<R> v) {
        return v.visitBool();
    }
}

