package ml2java.util;

import java.util.AbstractCollection;
import java.util.Arrays;
import java.util.Iterator;
import java.util.List;

// The first part of this file implements the Collection interface
// The second part of the file is generated by ml2java from the definition
//
//		type 'a simpleList =
//		  | Nil
//		  | Cons of 'a * 'a simpleList

public abstract class SimpleList<A> extends AbstractCollection<A> {

	public interface SimpleListVisitor<R, A> {

		public R visitNil();

		public R visitCons(A a, SimpleList<A> s);

	}

	public static <A> Integer size(SimpleList<A> l) {
		return l.accept(new SimpleListVisitor<Integer, A>() {

			@Override
			public Integer visitNil() {
				return 0;
			}

			@Override
			public Integer visitCons(A a, SimpleList<A> s) {
				return 1 + size(s);
			}
		});
	}

	public int size() {
		return size(this);
	}

	public static <A> Boolean isEmpty(SimpleList<A> l) {
		return l.size() == 0;
	}

	public static <A> A hd(SimpleList<A> l) {
		return l.accept(new SimpleListVisitor<A, A>() {

			@Override
			public A visitNil() {
				throw new Error("Match_failure");
			}

			@Override
			public A visitCons(A a, SimpleList<A> s) {
				return a;
			}
		});
	}

	public static <A> SimpleList<A> tl(SimpleList<A> l) {
		return l.accept(new SimpleListVisitor<SimpleList<A>, A>() {

			@Override
			public SimpleList<A> visitNil() {
				throw new Error("Match_failure");
			}

			@Override
			public SimpleList<A> visitCons(A a, SimpleList<A> s) {
				return s;
			}
		});
	}

	public static <A> Iterator<A> iterator(final SimpleList<A> l) {
		return new Iterator<A>() {

			private SimpleList<A> current = l;

			@Override
			public boolean hasNext() {
				return !isEmpty(current);
			}

			@Override
			public A next() {
				A h = hd(current);
				current = tl(current);
				return h;
			}
		};
	}

	@Override
	public Iterator<A> iterator() {
		return iterator(this);
	}

	public static <A> SimpleList<A> concat(final SimpleList<A> l1,
			final SimpleList<A> l2) {
		return l1.accept(new SimpleListVisitor<SimpleList<A>, A>() {

			@Override
			public SimpleList<A> visitNil() {
				return l2;
			}

			@Override
			public SimpleList<A> visitCons(A h, SimpleList<A> t) {
				return Cons(h, concat(t, l2));
			}
		});
	}

	/*
	 * public SimpleList<A> addAll(final SimpleList<A> l) { return concat(this,
	 * l); }
	 * 
	 * public SimpleList<A> add(final A a) { return Cons(a, this); }
	 */
	public static <A> SimpleList<A> fromList(List<A> l) {
		if (l.isEmpty()) {
			return Nil();
		} else {
			A hd = l.get(0);
			List<A> tl = l.subList(1, l.size());
			return Cons(hd, fromList(tl));
		}
	}

	public static <A> SimpleList<A> fromArray(A[] a) {
		return fromList(Arrays.asList(a));
	}

	public static <A> SimpleList<A> single(A a) {
		return Cons(a, Nil());
	}

	@SafeVarargs
	public static <A> SimpleList<A> list(A... a) {
		return fromList(Arrays.asList(a));
	}

	public static <A,R> R match(SimpleList<A> l, Function0<R> f1,
			Function2<A, SimpleList<A>, R> f2) {
		return l.match(f1, f2);
	}

	// The remaining of the file is auto-generated (do not modify)

	public abstract <R> R match(Function0<R> f1,
			Function2<A, SimpleList<A>, R> f2);

	public abstract <R> R accept(SimpleListVisitor<R, A> v);

	public static <A> SimpleList<A> Nil() {
		return new Nil<A>();
	}

	public static <A> SimpleList<A> Cons(A field1, SimpleList<A> field2) {
		return new Cons<A>(field1, field2);
	}

	public static <A> SimpleList<A> Cons2(SimpleList<A> field2, A field1) {
		return new Cons<A>(field1, field2);
	}

}

class Nil<A> extends SimpleList<A> {
	public Nil() {
	}

	@Override
	public int hashCode() {
		return ("").hashCode();
	}

	@Override
	public boolean equals(Object other) {
		if (other == null)
			return false;
		return (this.getClass() == (other.getClass()));
	}

	public <R> R match(Function0<R> f1, Function2<A, SimpleList<A>, R> f2) {
		return f1.apply();
	}

	public <R> R accept(SimpleListVisitor<R, A> v) {
		return v.visitNil();
	}
}

class Cons<A> extends SimpleList<A> {
	final A field1;
	final SimpleList<A> field2;
	final int code;

	public Cons(A field1, SimpleList<A> field2) {
		this.field1 = field1;
		this.field2 = field2;
		this.code = (Integer.toString(field1.hashCode()) + 
				Integer.toString(field2.hashCode())).hashCode();
	}

	@SuppressWarnings("unchecked")
	@Override
	public boolean equals(Object other) {
		if (other == null)
			return false;
		return (this.getClass() == (other.getClass())
				&& this.field1.equals(((Cons<A>) other).field1) && this.field2
					.equals(((Cons<A>) other).field2));
	}

	public <R> R match(Function0<R> f1, Function2<A, SimpleList<A>, R> f2) {
		return f2.apply(field1, field2);
	}

	public <R> R accept(SimpleListVisitor<R, A> v) {
		return v.visitCons(field1, field2);
	}
}
